//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

use crate::s5;

pub struct Window {
    size: s5::Rect,
    entities: Vec<Entity>,
}

impl Window {
    pub fn new(size: s5::Rect) -> Self {
        Self {
            size,
            entities: Vec::new(),
        }
    }

    pub fn size(&self) -> &s5::Rect {
        &self.size
    }

    pub fn resize(
        &mut self,
        size: s5::Rect,
    ) {
        self.size = size;
    }

    pub fn add_entity(
        &mut self,
        mut entity: Entity,
    ) {
        entity.stay_inside(&self.size);
        self.entities.push(entity);
    }

    pub fn redraw(
        &self,
        screen: &mut s5::Screen,
    ) {
        screen.clear(s5::Color {
            r: 100,
            g: 100,
            b: 100,
        });
        for entity in self.entities.iter() {
            entity.draw(screen);
        }
    }

    pub fn animate(
        &mut self,
        dt: f64,
    ) -> bool {
        for entity in self.entities.iter_mut() {
            entity.animate(dt);
            entity.stay_inside(&self.size);
        }
        true // go on
    }

    pub fn click(
        &mut self,
        mouse: s5::Point,
    ) {
        println!("click at {:?}", mouse);
        for entity in self.entities.iter_mut() {
            entity.click(&mouse);
        }
    }
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

pub struct Entity {
    kinetics: s5::KineticState,
    color: s5::Color,
    side: f64,
    bounding_radius: f64,
}

impl Entity {
    pub fn new(
        kinetics: s5::KineticState,
        color: s5::Color,
        side: f64,
    ) -> Self {
        let bounding_radius = 0.5 * std::f64::consts::SQRT_2 * side;
        Self {
            kinetics,
            color,
            side,
            bounding_radius,
        }
    }

    fn draw(
        &self,
        screen: &mut s5::Screen,
    ) {
        let point = s5::Point {
            x: self.kinetics.x,
            y: self.kinetics.y,
        };
        let size = s5::Rect {
            w: self.side,
            h: self.side,
        };
        let inner_size = s5::Rect {
            w: size.w * 0.75,
            h: size.h * 0.75,
        };
        screen.rectangle(
            &point,
            &size,
            Some(&inner_size),
            self.kinetics.angle,
            self.color,
        );
    }

    fn animate(
        &mut self,
        dt: f64,
    ) {
        self.kinetics.integrate(dt);
    }

    fn stay_inside(
        &mut self,
        window_size: &s5::Rect,
    ) {
        let kinetics = &mut self.kinetics;
        kinetics.x = kinetics.x.rem_euclid(window_size.w);
        kinetics.y = kinetics.y.rem_euclid(window_size.h);
    }

    fn click(
        &mut self,
        mouse: &s5::Point,
    ) {
        let radius = self.bounding_radius;
        let dx = mouse.x - self.kinetics.x;
        let dy = mouse.y - self.kinetics.y;
        if dx * dx + dy * dy <= radius * radius {
            let x_speed = self.kinetics.x_speed;
            let y_speed = self.kinetics.y_speed;
            self.kinetics.x_speed = -y_speed;
            self.kinetics.y_speed = x_speed;
            self.kinetics.angular_speed *= -1.0;
        }
    }
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
